# Autogenerated by Astropy's setup.py on 2020-04-02 01:20:13 UTC
import datetime

version = "4.0.1.post1"
githash = "bcb5b254c0f8723001b7be021159b07a40b5d9b2"


major = 4
minor = 0
bugfix = 1

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2020, 4, 2, 1, 20, 13)
debug = False

astropy_helpers_version = "4.0.1"
